'''
Indexing code for the KID search engine

Implemented for the reuter dataset

Indexes words under <title> and <text> tags.

Assumptions:
- A word is denoted as a string of letters
- HTML entities (e.g. &quot;) and tags (e.g.<p>) are removed
- All punctuation/underscores are collapsed/imploded (e.g. term_rank ->
  termrank)
- Words are stemmed using the porter stemmer
- Words appearing in the title are worth 10 normal words
'''

import re
import simplejson as json
import anydbm as dbm
import sha, time
from os import times

from Stemmer import Stemmer

class Document(object):
    '''
    A document class to be used in the storage of the forward keyword index
    '''

    kwdef = {'id':0, 'offset':-1, 'length':0, 'title':''}

    def __init__(self, **kwargs):
        for attr,val in self.kwdef.iteritems():
            if attr in kwargs:
                val = kwargs[attr]
            self.__setattr__(attr, val)

    def __len__(self):
        return self.length

    def serialize(self):
        return {
            'id'        : self.id,
            'offset'    : self.offset,
            'length'    : self.length,
            'title'     : self.title,
        }

    @classmethod
    def json_decode(cls, obj):
        for key in cls.kwdef:
            if not key in obj:
                return obj
            else:
                # Hack for json decoder (unicode to ascii)
                val = obj[key]
                del obj[key]
                if isinstance(val, unicode):
                    val = str(val)
                obj[key] = val
        return cls(**obj)

    def get_beef(self, origfile):
        '''
        Gets the full article from the `origfile` File object created
        by opening the file to which this document is attached
        '''
        origfile.seek(self.offset)
        return origfile.read(len(self))


class QueryEngine(object):
    ''' KiD (Keyword Index) Query Engine '''
    def __init__(self, dataset, forward, inverted):
        '''
        Initializes the kid QueryEngine
        `dataset`, `forward` and `inverted` filenames
        '''
        self.dataset = open(dataset)
        self.ddb = dbm.open(forward)
        self.tdb = dbm.open(inverted)

    def __del__(self):
        self.dataset.close()
        self.ddb.close()
        self.tdb.close()

    def run_query(self, qterms, not_in_index = []):
        '''
        Returns the intersection of documents linked to the `qterms`
        after stemming.

        Words that have not been indexed in processed form are appended
        in their lowercase form to `not_in_index`.
        '''
        qterms = list(set([x.lower() for x in qterms]) - stop_words)
        terms = stemmer.stemWords(qterms)
        if not terms:
            return set() 
        psets = []
        nsets = []
        for term in terms:
            sets = psets
            if term[0] == '~':
                sets = nsets
                term = term[1:]
            try:
                sets.append(set(x for x in json.loads(self.tdb[str(term)])))
            except KeyError:
                not_in_index.append(qterms[terms.index(term)])
                sets.append(set())
        if not (psets or nsets):
            return set()
        elif not psets:
            result_set = set(self.ddb.iterkeys())
        else:
            result_set = psets[0]
        for s in psets[1:]:
            result_set.intersection_update(s)
        for s in nsets:
            result_set -= s
        return result_set

    @staticmethod
    def parse_query(bool_query):
        '''
        Parses a boolean KiD query, returning a list of queries.

        `bool_query` is a preprocessed version of the original query of
        type list. Where each word is an element of the list, with the
        exception of the reserved words 'OR' and 'NOT'. Every two words
        are considered to be joined by an implicit 'AND'. 'NOT' is
        unirary operator with highest precedence, 'AND' and 'OR' have
        the same precedence.

        This parser does not consider bracketed queries.
        '''
        or_flag = False
        not_flag = False
        queries = [[]]
        for p in bool_query:
            if not p:
                continue
            if p == "OR":
                or_flag = True
            elif p == "NOT":
                not_flag = True
            elif or_flag:
                or_flag = False
                queries.append([p])
            else:
                if not_flag:
                    not_flag = False
                    p = '~%s' %p
                for q in queries:
                    q.append(p)
        return queries

    def search_query(self, bool_query, not_in_index=[]):
        '''
        Searches the `bool_query` and returns a set of satisfying docs

        `not_in_index` is updated with words were not found in
        our index in after all forms of pre-processing.
        '''
        qs = self.parse_query(bool_query)
        result_set = self.run_query(qs[0], not_in_index)
        for q in qs[1:]:
            result_set.update(self.run_query(q, not_in_index))
        return result_set

    def retrieve_doc(self, docid):
        '''
        Returns the Document object given the docid
        '''
        return json.loads(self.ddb[str(docid)], object_hook=Document.json_decode)

    def retrieve_doctext(self, doc):
        '''
        Returns the document text.
        `doc` can be a document object or a docid
        '''
        if isinstance(doc, (str, int)):
            doc = self.retrieve_doc(str(doc))
        return doc.get_beef(self.dataset)


class Timer(object):
    '''
    Class used to time the execution of code using os.times
    '''
    def __init__(self):
        self.full_reset()

    def full_reset(self):
        '''
        Does a reset of the totals as well as the counters
        '''
        self.reset()
        self._tu, self._ts, self._tr = [self._u , self._s, self._r]

    def reset(self):
        '''
        Reset the timer
        '''
        t = times()
        self._u , self._s, self._r = [t[0], t[1], t[4]]

    def stamp(self):
        '''
        Return the elapsed user, sys, real time since a reset|stamp
        and resets the timer
        '''
        u, s, r = [self._u, self._s, self._r]
        self.reset()
        u, s, r = [
          [self._u, self._s, self._r][i] - [u, s, r][i] for i in range(3)
        ]
        return "real = %.2f    user = %.2f    system=%.2f" %(r, u, s)

    def totals(self):
        '''
        Returns the total time elapsed since
        construction or full_reset.
        '''
        self.reset()
        u, s, r = [self._u, self._s, self._r]
        u, s, r = [
          [self._u, self._s, self._r][i] - [self._tu, self._ts, self._tr][i] \
          for i in range(3)
        ]
        return "real = %.2f    user = %.2f    system=%.2f" %(r, u, s)


class KidCursor(object):
    '''
    KiD Result Set Cursor
    '''
    def __init__(self, qe):
        '''
        Initializes a KiD cursor

        `qe` is a reference to a QueryEngine Instance
        '''
        self._qe = qe
        self._i = 0
        self.res = None
        self.count = 0
        self.time = None
        self.query = []
        self.unindexed = []

    def execute(self, bool_query):
        '''
        Executes a bool_query on the index, returning the number of
        results.
        '''
        self.query = bool_query
        t = Timer()
        self.unindexed = []
        self.res = list(self._qe.search_query(bool_query, self.unindexed))
        self.time = t.stamp()
        self.count = len(self.res)
        return len(self)

    def get_doctext(self, doc):
        ''' returns the text associated with doc.

        `doc` can be integer index of doc in result set or an actually
        indexer.Document object.
        '''
        if isinstance(doc, int):
            doc = self[doc]
        return self._qe.retrieve_doctext(doc)

    def __len__(self):
        return self.count

    def next(self):
        if self._i >= self.count:
            self._i = 0
            raise StopIteration
        else:
            c = self._i
            self._i += 1
            return self[c]

    def __getitem__(self, i):
        if isinstance(i, (int, long)):
            if 0 <= i < len(self):
                return self._qe.retrieve_doc(self.res[i])
            else:
                raise IndexError, ( "KidCursor[%d] out of range." % i )
        elif isinstance(i, slice):
            start, stop, step = i.start, i.stop, i.step
            step = 1 if not step else step
            if not start:
                start = 0
            start = start if start >=0 else len(self) + start
            if not stop:
                stop = len(self)
            stop = stop if stop >0 else len(self) + stop
            res = []
            for i in range(start, stop, step):
                try:
                    res.append(self[i])
                except IndexError:
                    break
            return res
        else:
            raise Exception("KidCursor received unsuported index type")


def engine_from_config(config, search_key="kid."):
    '''
    Returns a QueryEngine instance based upon the config dict locating
    fields using the search_key
    '''
    params = ['dataset', 'index.forward', 'index.inverted']
    values = [None, None, None]
    for key in params:
        if '%s%s' %(search_key, key) in config:
            values[params.index(key)] = config['%s%s' %(search_key, key)]
    return QueryEngine(*values)

def json_serialize(obj):
    if hasattr(obj, 'serialize'):
        obj = obj.serialize()
    return obj

def generate_indexs(input, forward, inverted):
    '''Generates the indices

    `input` filename containing the reuters dataset
    `forward` filename to store the forward dbm (docid -> docprop)
    `inverted` filename to store the inverted dbm (term -> docs)
    '''
    infile = open(input)
    ddb = dbm.open(forward, 'n')
    tdb = dbm.open('%s.stage1' %inverted, 'n')
    doc_id = 0
    state = False
    oline = infile.readline()
    while oline:
        line = oline.lower()
        if '<newsitem ' == line[:10]:
            state = True
            doc_id += 1
            doc = Document(
                id=doc_id,
                offset=infile.tell() - len(line),
                length = 0
            )
            words = {}
        if state:
            doc.length += len(line)
            if '</newsitem>' == line[:11]:
                state = False
                ddb[str(doc_id)] = json.dumps(doc, separators=(',', ':'), default=json_serialize)
                for term, weight in words.iteritems():
                    try:
                        tdb[term] += ',(%s,%s)' %(doc_id, weight)
                    except:
                        tdb[term] = '(%s,%s)' %(doc_id, weight)
            else:
                for x in '<title>','<p>':
                    if x in line:
                        if x == '<title>':
                            doc.title = oline.strip()
                        weight = WEIGHTS.get(x, WEIGHTS['default'])
                        line = punct.sub('', tags.sub(' ',line))
                        for word in stemmer.stemWords([
                            x.strip() for x in splitter.split(line)
                            if x not in stop_words
                        ]):
                            if word:
                                try:
                                    words[word] += weight
                                except KeyError:
                                    words[word] = weight
        oline = infile.readline()
    ddb.close()
    infile.close()

    idb = dbm.open(inverted, 'n')
    for term, docs in tdb.iteritems():
        idb[term] = json.dumps(dict(eval('(%s,)' %docs)), separators=(',', ':'))
    tdb.close()
    idb.close()

def make_cursor_id():
    return sha.new(str(time.time())).hexdigest()

def is_cursor(id):
    return id in __cursors

def get_cursor(duck):
    global __cursors
    if isinstance(duck, str):
        return __cursors[duck]
    cid = make_cursor_id()
    __cursors[cid] = KidCursor(duck)
    return cid

## Globals
__cursors = {}
WEIGHTS = {'<title>' : 10, 'default': 1}
tags = re.compile('</?[^>]+>|&[^;]+')
punct = re.compile('([^\w\s]|_)+')
splitter = re.compile('[^a-zA-Z]+')
stemmer = Stemmer('porter')
reserved_words = ["OR", "NOT"]
stop_words = set(("a", "about", "above", "across", "after",
"afterwards", "again", "against", "all", "almost", "alone", "along",
"already", "also", "although", "always", "am", "among", "amongst",
"amoungst", "amount", "an", "and", "another", "any", "anyhow", "anyone",
"anything", "anyway", "anywhere", "are", "around", "as", "at", "back",
"be", "became", "because", "become", "becomes", "becoming", "been",
"before", "beforehand", "behind", "being", "below", "beside", "besides",
"between", "beyond", "bill", "both", "bottom", "but", "by", "call",
"can", "cannot", "cant", "co", "computer", "con", "could", "couldnt",
"cry", "de", "describe", "detail", "do", "done", "down", "due",
"during", "each", "eg", "eight", "either", "eleven", "else",
"elsewhere", "empty", "enough", "etc", "even", "ever", "every",
"everyone", "everything", "everywhere", "except", "few", "fifteen",
"fify", "fill", "find", "fire", "first", "five", "for", "former",
"formerly", "forty", "found", "four", "from", "front", "full",
"further", "get", "give", "go", "had", "has", "hasnt", "have", "he",
"hence", "her", "here", "hereafter", "hereby", "herein", "hereupon",
"hers", "herself", "him", "himself", "his", "how", "however", "hundred",
"i", "ie", "if", "in", "inc", "indeed", "interest", "into", "is", "it",
"its", "itself", "keep", "last", "latter", "latterly", "least", "less",
"ltd", "made", "many", "may", "me", "meanwhile", "might", "mill",
"mine", "more", "moreover", "most", "mostly", "move", "much", "must",
"my", "myself", "name", "namely", "neither", "never", "nevertheless",
"next", "nine", "no", "nobody", "none", "noone", "nor", "not",
"nothing", "now", "nowhere", "of", "off", "often", "on", "once", "one",
"only", "onto", "or", "other", "others", "otherwise", "our", "ours",
"ourselves", "out", "over", "own", "part", "per", "perhaps", "please",
"put", "rather", "re", "same", "see", "seem", "seemed", "seeming",
"seems", "serious", "several", "she", "should", "show", "side", "since",
"sincere", "six", "sixty", "so", "some", "somehow", "someone",
"something", "sometime", "sometimes", "somewhere", "still", "such",
"system", "take", "ten", "than", "that", "the", "their", "them",
"themselves", "then", "thence", "there", "thereafter", "thereby",
"therefore", "therein", "thereupon", "these", "they", "thick", "thin",
"third", "this", "those", "though", "three", "through", "throughout",
"thru", "thus", "to", "together", "too", "top", "toward", "towards",
"twelve", "twenty", "two", "un", "under", "until", "up", "upon", "us",
"very", "via", "was", "we", "well", "were", "what", "whatever", "when",
"whence", "whenever", "where", "whereafter", "whereas", "whereby",
"wherein", "whereupon", "wherever", "whether", "which", "while",
"whither", "who", "hoever", "whole", "whom", "whose", "why", "will",
"with", "within", "without", "would", "yet", "you", "your", "yours",
"yourself", "yourselves"))
